/*   ArgoCraft Video Capture Board    */
/*   Driver and Sample Applications   */
/* Version 1.32 for Linux and FreeBSD */
/* Copyright (C) 1996 ArgoCraft Corp. */

#ifndef __KERNEL__
#define __KERNEL__
#endif /* __KERNEL__ */

#ifndef MODULE
#define MODULE
#endif /* MODULE */

#include <linux/bios32.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/stddef.h>
#include <linux/version.h>

#include "acvc.h"

#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif /* min */

#define VENDOR_ID 0x8086
#define DEVICE_ID 0x1223

#define LOBYTE(w) ((unsigned char) (unsigned short) (w))
#define HIBYTE(w) ((unsigned char) ((unsigned short) (w) >> 8))

#define LOWORD(dw) ((unsigned short) (unsigned long) (dw))
#define HIWORD(dw) ((unsigned short) ((unsigned long) (dw) >> 16))

#define MAXMEMTRYNUM 16
#define MAXRETRYNUM 8

#define alloc_32_pages() (__get_free_pages (GFP_KERNEL, 5, 0xffffffff))
#define free_32_pages(p) (free_pages (p, 5))

#define ACVC "acvc"
#define ACVC_MAJOR 60

#define MAX_WIDTH 640
#define MAX_HEIGHT 240

static unsigned char bus, devfun;
static volatile char * framebuf [2];

static unsigned short insize [2] = { MAX_WIDTH, MAX_HEIGHT };
static unsigned short exsize [2] = { MAX_WIDTH, MAX_HEIGHT };

static unsigned long fieldsize = 0, framesize = 0;
static unsigned long rest = 0;

static int incolor = 1; /* 0 == rgb16, 1 == rgb32 */
static int excolor = 0; /* 0 == color, 1 == mono */

#define DEF_BRI 0x8b
#define DEF_CON 0x46
#define DEF_SAT 0x40

static unsigned char src = 0;

static unsigned char bri = 0x8b;
static signed char con = 0x46;
static signed char sat = 0x40;

static int busy = 0;
static int pending = 0;
static int turn = 1;

static volatile unsigned long * memreg;

#define NUM_REG_7186 0x11

static unsigned char reg7186 [NUM_REG_7186] = {
  0x12, 0x40, 0x40, 0x10, 0x85, 0x78, 0x78, 0x10,
  0x80, 0x03, 0x80, 0x00, 0x00, 0xff, 0x00, 0x00,
  0x0f
};

#define NUM_REG_7110 0x35

static unsigned char reg7110 [NUM_REG_7110] = { 
  0x4c, 0x3c, 0x0d, 0xef, 0xbd, 0xf0, 0x00, 0x00,
  0xf8, 0xf8, 0x60, 0x60, 0x00, 0x06, 0x18, 0x90,
  0x00, 0x2c, 0x40, 0x46, 0x42, 0x1a, 0xff, 0xda,
  0xf0, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xd9, 0x16, 0x40, 0x41, 0x80, 0x41, 0x80, 0x4f,
  0xfe, 0x01, 0xcf, 0x0f, 0x03, 0x01, 0x81, 0x03,
  0x44, 0x75, 0x01, 0x8c, 0x03
};

#define NUM_SRC 6
#define NUM_SEL 8

static const unsigned char srcsel [NUM_SEL] = {
  0x06, 0x20, 0x21, 0x22, 0x2c, 0x30, 0x31, 0x21
};

static const unsigned char srcval [NUM_SRC] [NUM_SEL] = {
  { 0x00, 0xd9, 0x17, 0x40, 0x03, 0x44, 0x75, 0x16 }, /* mode 0 */
  { 0x00, 0xd8, 0x17, 0x40, 0x03, 0x44, 0x75, 0x16 }, /* mode 1 */
  { 0x00, 0xba, 0x07, 0x91, 0x03, 0x60, 0xb5, 0x05 }, /* mode 2 */
  { 0x00, 0xb8, 0x07, 0x91, 0x03, 0x60, 0xb5, 0x05 }, /* mode 3 */
  { 0x00, 0x7c, 0x07, 0xd2, 0x83, 0x60, 0xb5, 0x03 }, /* mode 4 */
  { 0x00, 0x78, 0x07, 0xd2, 0x83, 0x60, 0xb5, 0x03 }, /* mode 5 */
};

static void longwait (void) {
  unsigned long limit;

  limit = jiffies + 10;

  while (jiffies < limit) {
    schedule ();
  }
}

static void shortwait (void) {
  unsigned long limit;

  limit = jiffies + 1;

  while (jiffies < limit) {
    schedule ();
  }
}

static unsigned long read7116 (unsigned char addr) {
  return memreg [addr / 4];
}

static void write7116 (unsigned char addr, unsigned long val) {
  memreg [addr / 4] = val;
}

static void write7186 (unsigned char addr, unsigned char val) {
  unsigned long status, count = 0, timeout = 10;

  write7116 (0x64, 0x01b80000 | addr << 8 | val);
  reg7186 [addr] = val;

 retry:

  while (read7116 (0x64) & 0x01000000) {
    shortwait ();

    if (! timeout --) {
      printk ("acvc: write7186 timed out\n");
      return;
    }
  }

  if ((status = read7116 (0x60)) & 0x00000020) {
    write7116 (0x60, status);

    if (++ count > MAXRETRYNUM) {
      printk ("acvc: too many retries\n");
      return;
    }

    /* printk ("acvc: retrying... (%ld)\n", count); */
    goto retry;
  }
}

static void write7110 (unsigned char addr, unsigned char val) {
  unsigned long status, count = 0, timeout = 10;

  write7116 (0x64, 0x019c0000 | addr << 8 | val);
  reg7110 [addr] = val;

 retry:

  while (read7116 (0x64) & 0x01000000) {
    shortwait ();

    if (! timeout --) {
      printk ("acvc: write7110 timed out\n");
      return;
    }
  }

  if ((status = read7116 (0x60)) & 0x00000020) {
    write7116 (0x60, status);

    if (++ count > MAXRETRYNUM) {
      printk ("acvc: too many retries\n");
      return;
    }

    /* printk ("acvc: retrying... (%ld)\n", count); */
    goto retry;
  }
}

static int setin (unsigned short x, unsigned short y, int c) {
  unsigned long size;

  if (! c && excolor) {
    return - 1;
  }

  if (! x || ! y) {
    return - 1;
  }

  if (x > exsize [0] || y > exsize [1]) {
    return - 1;
  }

  if (! c && x % 2) {
    return - 1;
  }

  size = x * y * (c ? 4 : 2);

  fieldsize = x * (c ? 4 : 2);
  framesize = size;

  rest = framesize;

  insize [0] = x;
  insize [1] = y;
  incolor = c;

  if (c) {
    write7116 (0x30, 0x39393900);
    write7116 (0x34, 0x39393900);
    write7186 (0x00, reg7186 [0x00] | 0x02);
  } else {
    write7116 (0x30, 0xeeeeee01);
    write7116 (0x34, 0xeeeeee01);
    write7186 (0x00, reg7186 [0x00] & 0xfd);
  }

  write7116 (0x8c, (unsigned long) framebuf [0] + size);
  write7116 (0x90, (unsigned long) framebuf [1] + size);

  write7186 (0x01, LOBYTE (x));
  write7186 (0x04, HIBYTE (x) | (reg7186 [0x04] & 0xfc));

  write7186 (0x05, LOBYTE (y));
  write7186 (0x08, HIBYTE (y) | (reg7186 [0x08] & 0xfc));

  return 0;
}

static int setexsize (unsigned short x, unsigned short y) {
  unsigned short ox, oy;

  if (x > MAX_WIDTH || y > MAX_HEIGHT) {
    return - 1;
  }

  if (x < insize [0] || y < insize [1]) {
    return - 1;
  }

  exsize [0] = x;
  exsize [1] = y;

  write7186 (0x02, LOBYTE (x));
  write7186 (0x04, HIBYTE (x) << 2 | (reg7186 [0x04] & 0xf3));

  write7186 (0x06, LOBYTE (y));
  write7186 (0x08, HIBYTE (y) << 2 | (reg7186 [0x08] & 0xf3));

  ox = 0x10 + ((MAX_WIDTH - x) / 2);
  oy = 0x10 + ((MAX_HEIGHT - y) / 2);
  
  write7186 (0x03, LOBYTE (ox));

  if (HIBYTE (ox) == 0) {
    write7186 (0x04, reg7186 [0x04] & 0xef);
  } else {
    write7186 (0x04, reg7186 [0x04] | 0x10);
  }

  write7186 (0x07, LOBYTE (oy));

  if (HIBYTE (oy) == 0) {
    write7186 (0x08, reg7186 [0x08] & 0xef);
  } else {
    write7186 (0x08, reg7186 [0x08] | 0x10);
  }

  return 0;
}

static int setexcolor (int e) {
  if (e & ! incolor) {
    return - 1;
  }

  excolor = e;

  if (e) {
    write7186 (0x00, reg7186 [0x00] | 0x01);
    write7186 (0x10, reg7186 [0x10] | 0x10);
  } else {
    write7186 (0x00, reg7186 [0x00] & 0xfe);
    write7186 (0x10, reg7186 [0x10] & 0xef);
  }

  return 0;
}

static int setsrc (unsigned char s) {
  int i;

  if (s >= NUM_SRC) {
    return - 1;
  }

  src = s;

  for (i = 0; i < NUM_SEL; i ++) {
    write7110 (srcsel [i], srcval [s] [i]);
  }

  return 0;
}

static void setbri (unsigned char b) {
  bri = b;
  write7110 (0x19, b);
}

static void setcon (signed char c) {
  con = c;
  write7110 (0x13, c);
}

static void setsat (signed char s) {
  sat = s;
  write7110 (0x12, s);
}

static void init7116 (void) {
  /* initialization */

  write7116 (0x60, 0);
  write7116 (0x40, 0);
  
  longwait ();
  
  write7116 (0x40, 0x00000040);
  
  longwait ();
  
  write7116 (0x00, 0x00000004);
  write7116 (0x04, 0x00000004);
  write7116 (0x08, 0x00000004);
  write7116 (0x0c, 0x00000004);
  write7116 (0x10, 0x00000004);
  write7116 (0x14, 0x00000004);
  write7116 (0x8c, 0);
  write7116 (0x90, 0);
  write7116 (0x5c, 0x80404020);
  write7116 (0x40, 0x000080c0);
  
  longwait ();
  
  write7116 (0x40, 0x00008040);
  
  longwait ();
  
  /* DMA */

  write7116 (0x00, (unsigned long) framebuf [0]);
  write7116 (0x04, 0);
  write7116 (0x08, 0);
  write7116 (0x0c, (unsigned long) framebuf [1]);
  write7116 (0x10, 0);
  write7116 (0x14, 0);
  
  /* Stride */

  write7116 (0x18, 0);
  write7116 (0x1c, 0);
  write7116 (0x20, 0);
  write7116 (0x24, 0);
  write7116 (0x28, 0);
  write7116 (0x2c, 0);

  /* Capture Control */

  write7116 (0x38, 0x00200020);
  write7116 (0x3c, 0x00000003);
  write7116 (0x40, 0x000000c0);
  write7116 (0x44, 0);
  write7116 (0x48, 0);
  write7116 (0x4c, 0x00000001);
  write7116 (0x50, 0x00000001);
  write7116 (0x54, 0);
  write7116 (0x58, 0x0005007c);
  
  /* I2C */

  write7116 (0x5c, 0x461e1e0f);
  write7116 (0x60, 0);
  write7116 (0x64, 0);
  write7116 (0x68, 0);
  write7116 (0x6c, 0);
  write7116 (0x70, 0);
  write7116 (0x74, 0);
  write7116 (0x78, 0);
  write7116 (0x7c, 0);
  write7116 (0x80, 0);
  write7116 (0x84, 0);
  write7116 (0x88, 0);
  
  write7116 (0x40, 0x00008ff0);
}

static void init7186 (void) {
  unsigned char addr = NUM_REG_7186;

  while (addr --) {
    write7186 (addr, reg7186 [addr]);
  }
}

static void init7110 (void) {
  unsigned char addr = NUM_REG_7110;

  while (addr --) {
    write7110 (addr, reg7110 [addr]);
  }
}

static int iscap (void) {
  return (read7116 (0x40) & 0x0000000c) ? 1 : 0;
}

static void capstart (int which) {
  const unsigned long val [3] = { 0x00008ffc, 0x00008ff4, 0x00008ff8 };

  if (which == 3) {
    which = turn;
  }

  write7116 (0x40, val [which]);
  pending = 1;
  rest = framesize;
}

static int capwait (int which) {
  const unsigned long val [3] = { 0x00000030, 0x00000010, 0x00000020 };
  unsigned long status, count = 0, timeout;

  if (which == 3) {
    which = turn;
  }

 retry:

  if (! pending) {
    return 0;
  }

  pending = 0;

  /* wait for capture */

  timeout = 100;

  while (iscap ()) {
    shortwait ();

    if (! timeout --) {
      printk ("acvc: capture timed out\n");
      return - 1;
    }
  }

  /* wait for transfer */

  timeout = 100;

  while (~ read7116 (0x40) & val [which]) {
    shortwait ();

    if (! timeout --) {
      printk ("acvc: transfer timed out\n");
      return - 1;
    }
  }

  /* check up errors */

  if ((status = read7116 (0x40)) & 0x00000f00) {
    write7116 (0x40, status);

    if (status & 0x00000400) {
      printk ("acvc: address error (even)\n");
    }
    
    if (status & 0x00000800) {
      printk ("acvc: address error (odd)\n");
    }

    if (status & 0x00000c00) {
      return - 1;
    }

    if (status & 0x00000100) {
      /* printk ("acvc: field corrupt (even)\n"); */
    }

    if (status & 0x00000200) {
      /* printk ("acvc: field corrupt (odd)\n"); */
    }

    if (status & 0x00000300) {
      if (++ count > MAXRETRYNUM) {
	printk ("acvc: too many retries\n");
	return - 1;
      }
      
      /* printk ("acvc: retrying... (%ld)\n", count); */
      capstart (which);
      goto retry;
    }
  }

  return 0;
}

static int acvc_lseek (struct inode * inode,
		       struct file * file,
		       off_t offset,
		       int origin) {
  return - ESPIPE;
}

static int acvc_read (struct inode * inode,
		      struct file * file,
		      char * buf,
		      int count) {
  unsigned int minor;
  char * p, * q;
  int i, ret, len;

  minor = MINOR (inode -> i_rdev);

  if (minor > 3) {
    return - ENODEV;
  }
  
  if (minor) {
    len = min (count, rest);

    if (len == 0) {
      return 0;
    }

    if ((ret = verify_area (VERIFY_WRITE, buf, len))) {
      return ret;
    }

    if (capwait (minor)) {
      return - EIO;
    }

    if (minor == 3) {
      memcpy_tofs (buf,
		   (char *) framebuf [turn - 1] + (framesize - rest),
		   len);
    } else {
      memcpy_tofs (buf,
		   (char *) framebuf [minor - 1] + (framesize - rest),
		   len);
    }

    rest -= len;

    return len;
  }

  if (count < framesize * 2) {
    return - EINVAL;
  }

  if ((ret = verify_area (VERIFY_WRITE, buf, framesize * 2))) {
    return ret;
  }

  p = (char *) framebuf [0];
  q = (char *) framebuf [1];

  if (capwait (minor)) {
    return - EIO;
  }

  for (i = 0; i < insize [1]; i ++) {
    memcpy_tofs (buf, p, fieldsize);

    p += fieldsize;
    buf += fieldsize;

    memcpy_tofs (buf, q, fieldsize);

    q += fieldsize;
    buf += fieldsize;
  }

  return framesize * 2;
}

static int acvc_ioctl (struct inode * inode, struct file * file,
		       unsigned int cmd, unsigned long arg) {
  unsigned int minor;

  minor = MINOR (inode -> i_rdev);

  switch (cmd) {
  case ACVC_ISCAP:
    {
      int ret, val;

      if ((ret = verify_area (VERIFY_WRITE, (void *) arg, sizeof (int)))) {
	return ret;
      }

      val = iscap ();
      memcpy_tofs ((int *) arg, & val, sizeof (int));

      return 0;
    }
  case ACVC_START:
    if (capwait (minor)) {
      return - EIO;
    }

    turn = 3 - turn;

    capstart (minor);

    return 0;
  case ACVC_WAIT:
    if (capwait (minor)) {
      return - EIO;
    }

    return 0;
  case ACVC_SIZE:
    {
      int c;

      c = (arg & 0x80000000) ? 1 : 0;
      arg &= 0x7fffffff;

      if (capwait (minor)) {
	return - EIO;
      }
      
      if (minor) {
	if (setin (HIWORD (arg), LOWORD (arg), c)) {
	  return - EINVAL;
	}
      } else {
	if (LOWORD (arg) % 2) {
	  return - EINVAL;
	}

	if (setin (HIWORD (arg), LOWORD (arg) / 2, c)) {
	  return - EINVAL;
	}
      }

      return 0;
    }
  case ACVC_VIEW:
    if (capwait (minor)) {
      return - EIO;
    }

    if (LOWORD (arg) % 2) {
      return - EINVAL;
    }

    if (setexsize (HIWORD (arg), LOWORD (arg) / 2)) {
      return - EINVAL;
    }

    return 0;
  case ACVC_MONO:
    if (capwait (minor)) {
      return - EIO;
    }

    if (setexcolor (arg)) {
      return - EINVAL;
    }

    return 0;
  case ACVC_SRC:
    if (capwait (minor)) {
      return - EIO;
    }

    if (setsrc (arg)) {
      return - EINVAL;
    }

    return 0;
  case ACVC_BRI:
    if (capwait (minor)) {
      return - EIO;
    }

    if (arg & 0xffffff00) {
      return - EINVAL;
    }

    setbri ((unsigned char) arg);

    return 0;
  case ACVC_CON:
    if (capwait (minor)) {
      return - EIO;
    }

    if (arg & 0xffffff00) {
      return - EINVAL;
    }

    setcon ((signed char) (unsigned char) arg);

    return 0;
  case ACVC_SAT:
    if (capwait (minor)) {
      return - EIO;
    }

    if (arg & 0xffffff00) {
      return - EINVAL;
    }

    setsat ((signed char) (unsigned char) arg);

    return 0;
  default:
    return - EINVAL;
  }
}

static int acvc_open (struct inode * inode,
		      struct file * file) {
  unsigned int minor;

  minor = MINOR (inode -> i_rdev);

  if (busy) {
    return - EBUSY;
  }

  busy = 1;
  MOD_INC_USE_COUNT;

  pending = 0;
  turn = 1;

  setin (MAX_WIDTH / (minor ? 2 : 1), MAX_HEIGHT, 1);
  setexsize (MAX_WIDTH, MAX_HEIGHT);
  setexcolor (0);

  setsrc (0);

  setbri (DEF_BRI);
  setcon (DEF_CON);
  setsat (DEF_SAT);

  capstart (minor);

  return 0;
}

static void acvc_release (struct inode * inode,
			  struct file * file) {
  unsigned int minor;

  minor = MINOR (inode -> i_rdev);
  capwait (minor);

  busy = 0;
  MOD_DEC_USE_COUNT;

  return;
}

static struct file_operations acvc_fops = {
  acvc_lseek,
  acvc_read,
  NULL, /* acvc_write */
  NULL, /* acvc_readdir */
  NULL, /* acvc_select */
  acvc_ioctl,
  NULL, /* acvc_mmap */
  acvc_open,
  acvc_release,
  NULL, /* acvc_fsync */
  NULL, /* acvc_fasync */
  NULL, /* acvc_check_media_change */
  NULL /* acvc_revalidate */
};

static char * alloc_buf (void) {
  unsigned long addrs [MAXMEMTRYNUM], addr, retval;
  int num, i, j, k;

  for (num = 0; num < MAXMEMTRYNUM; num ++) {
    if ((addrs [num] = alloc_32_pages ()) == 0) {
      break;
    }
  }

/*
  for (i = 0; i < num; i ++) {
    printk ("acvc: addrs [%i] = %.8X\n", i, (int) addrs [i]);
  }
*/

  for (i = 0; i < num; i ++) {
    for (j = 0; j < 5; j ++) {
      addr = addrs [i] + j * 32 * PAGE_SIZE;
      for (k = 0; k < num; k ++) {
	if (addr == addrs [k]) {
	  goto good;
	}
      }
      goto bad;
    good: ;
    }
    retval = addrs [i];
    goto success;
  bad: ;
  }
  goto failure;

 failure: ;

  for (i = 0; i < num; i ++) {
    free_32_pages (addrs [i]);
  }

  return NULL;

 success: ;

  for (i = 0; i < 5; i ++) {
    addr = retval + i * 32 * PAGE_SIZE;
    for (j = 0; j < num; j ++) {
      if (addr == addrs [j]) {
	addrs [j] = 0;
      }
    }
  }

  for (k = 0; k < num; k ++) {
    if (addrs [k]) {
      free_32_pages (addrs [k]);
    }
  }

  return (char *) retval;
}

static void free_buf (volatile char * buf) {
  int i;

  for (i = 0; i < 5; i ++) {
    free_32_pages ((unsigned long) buf + i * 32 * PAGE_SIZE);
  }
}

int init_module (void) {
  static unsigned long memreg_;

  printk ("acvc: version 1.32\n");

  if (pcibios_find_device (VENDOR_ID, DEVICE_ID, 0, & bus, & devfun)) {
    printk ("acvc: can't find device\n");
    return - ENODEV;
  }

  if ((framebuf [0] = alloc_buf ()) == NULL) {
    printk ("acvc: can't allocate physical memory (even)\n");
    return - ENOMEM;
  }

  printk ("acvc: framebuf (even) == %p\n", framebuf [0]);

  if ((framebuf [1] = alloc_buf ()) == NULL) {
    free_buf (framebuf [0]);

    printk ("acvc: can't allocate physical memory (odd)\n");
    return - ENOMEM;
  }

  printk ("acvc: framebuf (odd) == %p\n", framebuf [1]);

  if (register_chrdev (ACVC_MAJOR, ACVC, & acvc_fops)) {
    free_buf (framebuf [0]);
    free_buf (framebuf [1]);

    printk ("acvc: can't get major number %d\n", ACVC_MAJOR);
    return - EBUSY;
  }

  /* enable master & memory */
  if (pcibios_write_config_byte (bus, devfun,
				 PCI_COMMAND,
				 PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER)) {
    unregister_chrdev (ACVC_MAJOR, ACVC);
    free_buf (framebuf [0]);
    free_buf (framebuf [1]);

    printk ("acvc: can't enable master and memory\n");
    return - EIO;
  }

  if (pcibios_read_config_dword (bus, devfun,
				 PCI_BASE_ADDRESS_0,
				 (unsigned int *) & memreg_)) {
    unregister_chrdev (ACVC_MAJOR, ACVC);
    free_buf (framebuf [0]);
    free_buf (framebuf [1]);

    printk ("acvc: can't get memory base address\n");
    return - EIO;
  }

  if ((memreg = vremap (memreg_, PAGE_SIZE)) == NULL) {
    unregister_chrdev (ACVC_MAJOR, ACVC);

    free_buf (framebuf [0]);
    free_buf (framebuf [1]);

    printk ("acvc: can't map memory registers\n");
    return - ENOMEM;
  }

  init7116 ();

  if (read7116 (0) != (unsigned long) framebuf [0]) {
    vfree ((void *) memreg);

    unregister_chrdev (ACVC_MAJOR, ACVC);

    free_buf (framebuf [0]);
    free_buf (framebuf [1]);

    printk ("acvc: init7116 failed\n");
    return - EIO;
  }

  init7186 ();
  init7110 ();

  setin (MAX_WIDTH, MAX_HEIGHT, 1);
  setexsize (MAX_WIDTH, MAX_HEIGHT);
  setexcolor (0);

  setsrc (0);

  setbri (DEF_BRI);
  setcon (DEF_CON);
  setsat (DEF_SAT);

  capstart (0);

  if (capwait (0)) {
    vfree ((void *) memreg);

    unregister_chrdev (ACVC_MAJOR, ACVC);

    free_buf (framebuf [0]);
    free_buf (framebuf [1]);

    printk ("acvc: capture failed\n");
    return - EIO;
  }

  printk ("acvc: initialized\n");

  return 0;
}

void cleanup_module (void) {
  busy = 1;
  capwait (0);

  vfree ((void *) memreg);

  if (unregister_chrdev (ACVC_MAJOR, ACVC)) {
    printk ("acvc: unregister_chrdev failed\n");
  }

  free_buf (framebuf [0]);
  free_buf (framebuf [1]);
}
